'use strict';

Object.defineProperty(exports, '__esModule', { value: true });

function _interopDefault (ex) { return (ex && (typeof ex === 'object') && 'default' in ex) ? ex['default'] : ex; }

var _construct = _interopDefault(require('@babel/runtime/helpers/construct'));
var _objectWithoutPropertiesLoose = _interopDefault(require('@babel/runtime/helpers/objectWithoutPropertiesLoose'));
var _extends = _interopDefault(require('@babel/runtime/helpers/extends'));
var THREE = require('three/src/Three');
var React = require('react');
var React__default = _interopDefault(React);
var Reconciler = _interopDefault(require('react-reconciler'));
var omit = _interopDefault(require('lodash-es/omit'));
var upperFirst = _interopDefault(require('lodash-es/upperFirst'));
var ResizeObserver = _interopDefault(require('resize-observer-polyfill'));
var scheduler = require('scheduler');

var roots = new Map();
var emptyObject = {};
var globalSubscribersMap = new Set();
var globalSubscribers = [];
var useFrameloop = function useFrameloop(fn) {
  globalSubscribersMap.add(fn);
  globalSubscribers = Array.from(globalSubscribersMap.values());
};
var catalogue = {};
var apply = function apply(objects) {
  return catalogue = _extends({}, catalogue, objects);
};
function applyProps(instance, newProps, oldProps, interpolateArray) {
  if (oldProps === void 0) {
    oldProps = {};
  }

  if (interpolateArray === void 0) {
    interpolateArray = false;
  }

  if (instance.obj) instance = instance.obj; // Filter equals, events and reserved props

  var sameProps = Object.keys(newProps).filter(function (key) {
    return newProps[key] === oldProps[key];
  });
  var handlers = Object.keys(newProps).filter(function (key) {
    return typeof newProps[key] === 'function' && key.startsWith('on');
  });
  var filteredProps = omit(newProps, [].concat(sameProps, ['children', 'key', 'ref']));

  if (Object.keys(filteredProps).length > 0) {
    Object.entries(filteredProps).forEach(function (_ref) {
      var key = _ref[0],
          value = _ref[1];

      if (!handlers.includes(key)) {
        var root = instance;
        var target = root[key];

        if (key.includes('-')) {
          var entries = key.split('-');
          target = entries.reduce(function (acc, key) {
            return acc[key];
          }, instance);

          if (target && !target.set) {
            // The target is atomic, this forces us to switch the root
            var _entries$reverse = entries.reverse(),
                name = _entries$reverse[0],
                reverseEntries = _entries$reverse.slice(1);

            root = reverseEntries.reverse().reduce(function (acc, key) {
              return acc[key];
            }, instance);
            key = name;
          }
        }

        if (target && target.set) {
          if (target.constructor.name === value.constructor.name) {
            target.copy(value);
          } else if (Array.isArray(value)) {
            var _target;

            (_target = target).set.apply(_target, value);
          } else {
            target.set(value);
          }
        } else {
          if (interpolateArray && Array.isArray(root[key])) root[key].push(value);else root[key] = value;
        }
      }
    }); // Prep interaction handlers

    if (handlers.length) {
      instance.__handlers = handlers.reduce(function (acc, key) {
        var _extends2;

        var name = key.charAt(2).toLowerCase() + key.substr(3);
        return _extends({}, acc, (_extends2 = {}, _extends2[name] = newProps[key], _extends2));
      }, {});
    } // Call the update lifecycle, if present


    if (instance.__handlers && instance.__handlers.update) instance.__handlers.update(instance);
  }
}

function createInstance(type, _ref2) {
  var _ref2$args = _ref2.args,
      args = _ref2$args === void 0 ? [] : _ref2$args,
      props = _objectWithoutPropertiesLoose(_ref2, ["args"]);

  var name = upperFirst(type);
  var instance;
  if (type === 'primitive') instance = props.object;else {
    var target = catalogue[name] || THREE[name];
    instance = Array.isArray(args) ? _construct(target, args) : new target(args);
  }
  if (!instance.isObject3D) instance = {
    obj: instance,
    parent: undefined
  };
  applyProps(instance, props, {});
  return instance;
}

function appendChild(parentInstance, child) {
  if (child) {
    if (child.isObject3D) parentInstance.add(child);else {
      var _applyProps;

      child.parent = parentInstance; // The name attribute implies that the object attaches itself on the parent

      if (child.obj.name) applyProps(parentInstance, (_applyProps = {}, _applyProps[child.obj.name] = child.obj, _applyProps), {}, true);
    }
  }
}

function removeChild(parentInstance, child) {
  if (child) {
    if (child.isObject3D) {
      parentInstance.remove(child);
      if (child.dispose) child.dispose();
    } else {
      child.parent = undefined;

      if (child.obj.name) {
        // Remove attachment
        var target = parentInstance[child.obj.name];
        if (Array.isArray(target)) parentInstance[child.obj.name] = target.filter(function (x) {
          return x !== child.obj;
        });else parentInstance[child.obj.name] = undefined;
      }

      if (child.obj.dispose) child.obj.dispose();
    }
  }
}

var Renderer = Reconciler({
  now: scheduler.unstable_now,
  createInstance: createInstance,
  removeChild: removeChild,
  appendChild: appendChild,
  supportsMutation: true,
  isPrimaryRenderer: false,
  schedulePassiveEffects: scheduler.unstable_scheduleCallback,
  cancelPassiveEffects: scheduler.unstable_cancelCallback,
  appendInitialChild: appendChild,
  appendChildToContainer: appendChild,
  removeChildFromContainer: removeChild,
  insertBefore: function insertBefore(parentInstance, child, beforeChild) {
    if (child) {
      if (child.isObject3D) {
        child.parent = parentInstance;
        child.dispatchEvent({
          type: 'added'
        }); // TODO: the order is out of whack if data objects are present, has to be recalculated

        var index = parentInstance.children.indexOf(beforeChild);
        parentInstance.children = [].concat(parentInstance.children.slice(0, index), [child], parentInstance.children.slice(index));
      } else child.parent = parentInstance;
    }
  },
  commitUpdate: function commitUpdate(instance, updatePayload, type, oldProps, newProps, fiber) {
    if (instance.isObject3D) {
      applyProps(instance, newProps, oldProps);
    } else {
      // This is a data object, let's extract critical information about it
      var parent = instance.parent;

      var _newProps$args = newProps.args,
          argsNew = _newProps$args === void 0 ? [] : _newProps$args,
          restNew = _objectWithoutPropertiesLoose(newProps, ["args"]);

      var _oldProps$args = oldProps.args,
          argsOld = _oldProps$args === void 0 ? [] : _oldProps$args,
          restOld = _objectWithoutPropertiesLoose(oldProps, ["args"]); // If it has new props or arguments, then it needs to be re-instanciated


      if (argsNew.some(function (value, index) {
        return value !== argsOld[index];
      })) {
        // Next we create a new instance and append it again
        var newInstance = createInstance(instance.obj.type, newProps);
        removeChild(parent, instance);
        appendChild(parent, newInstance); // Switch instance

        instance.obj = newInstance.obj;
        instance.parent = newInstance.parent;
      } else {
        // Otherwise just overwrite props
        applyProps(instance, restNew, restOld);
      }
    }
  },
  getPublicInstance: function getPublicInstance(instance) {
    return instance;
  },
  getRootHostContext: function getRootHostContext(rootContainerInstance) {
    return emptyObject;
  },
  getChildHostContext: function getChildHostContext(parentHostContext, type) {
    return emptyObject;
  },
  createTextInstance: function createTextInstance() {},
  finalizeInitialChildren: function finalizeInitialChildren(instance, type, props, rootContainerInstance) {
    return false;
  },
  prepareUpdate: function prepareUpdate(instance, type, oldProps, newProps, rootContainerInstance, hostContext) {
    return emptyObject;
  },
  shouldDeprioritizeSubtree: function shouldDeprioritizeSubtree(type, props) {
    return false;
  },
  prepareForCommit: function prepareForCommit() {},
  resetAfterCommit: function resetAfterCommit() {},
  shouldSetTextContent: function shouldSetTextContent(props) {
    return false;
  }
});
function render(element, container) {
  var root = roots.get(container);

  if (!root) {
    root = Renderer.createContainer(container);
    roots.set(container, root);
  }

  Renderer.updateContainer(element, root, null, undefined);
  return Renderer.getPublicRootInstance(root);
}
function unmountComponentAtNode(container) {
  var root = roots.get(container);
  if (root) Renderer.updateContainer(null, root, null, function () {
    return roots.delete(container);
  });
}

function useMeasure() {
  var ref = React.useRef();

  var _useState = React.useState({
    left: 0,
    top: 0,
    width: 0,
    height: 0
  }),
      bounds = _useState[0],
      set = _useState[1];

  var _useState2 = React.useState(function () {
    return new ResizeObserver(function (_ref3) {
      var entry = _ref3[0];
      return set(entry.contentRect);
    });
  }),
      ro = _useState2[0];

  React.useEffect(function () {
    if (ref.current) ro.observe(ref.current);
    return function () {
      return ro.disconnect();
    };
  }, []);
  return [{
    ref: ref
  }, bounds];
}

var context = React__default.createContext();
var Canvas = React__default.memo(function (_ref4) {
  var children = _ref4.children,
      props = _ref4.props,
      style = _ref4.style,
      camera = _ref4.camera,
      renderFn = _ref4.render,
      resize = _ref4.resize,
      created = _ref4.created,
      rest = _objectWithoutPropertiesLoose(_ref4, ["children", "props", "style", "camera", "render", "resize", "created"]);

  var canvas = React.useRef();
  var state = React.useRef({
    subscribers: [],
    render: true,
    active: true,
    canvas: undefined,
    gl: undefined,
    camera: undefined,
    scene: undefined,
    size: undefined,
    viewport: function viewport(target) {
      if (target === void 0) {
        target = new THREE.Vector3(0, 0, 0);
      }

      var distance = state.current.camera.position.distanceTo(target);
      var fov = THREE.Math.degToRad(state.current.camera.fov); // convert vertical fov to radians

      var height = 2 * Math.tan(fov / 2) * distance; // visible height

      var width = height * state.current.camera.aspect;
      return {
        width: width,
        height: height
      };
    },
    subscribe: function subscribe(fn, main) {
      if (main !== void 0) state.current.render = main ? false : true;
      state.current.subscribers.push(fn);
      return function () {
        return state.current.subscribers = state.current.subscribers.filter(function (s) {
          return s === fn;
        });
      };
    }
  });

  var _useMeasure = useMeasure(),
      bind = _useMeasure[0],
      size = _useMeasure[1];

  state.current.size = size;

  var _useState3 = React.useState(false),
      ready = _useState3[0],
      setReady = _useState3[1];

  var readyRef = React.useRef(false);
  React.useEffect(function () {
    return void (readyRef.current = ready);
  }, [ready]);

  var _useState4 = React.useState(function () {
    return new THREE.Raycaster();
  }),
      raycaster = _useState4[0];

  var _useState5 = React.useState(function () {
    return new THREE.Vector2();
  }),
      mouse = _useState5[0];

  var _useState6 = React.useState('default'),
      cursor = _useState6[0],
      setCursor = _useState6[1];

  React.useEffect(function () {
    state.current.gl = new THREE.WebGLRenderer(_extends({
      canvas: canvas.current,
      antialias: true,
      alpha: true
    }, props));
    state.current.gl.setClearAlpha(0);
    state.current.gl.setSize(state.current.size.width, state.current.size.height, false); //state.current.gl.setPixelRatio(window.devicePixelRatio)
    //state.current.gl.setSize(window.innerWidth, window.innerHeight)

    state.current.camera = camera && camera.current || new THREE.PerspectiveCamera(75, 0, 0.1, 1000);
    state.current.camera.position.z = 5;
    state.current.canvas = canvas.current;
    state.current.scene = window.scene = new THREE.Scene();

    var renderLoop = function renderLoop() {
      if (!state.current.active) return;
      requestAnimationFrame(renderLoop);

      if (readyRef.current) {
        globalSubscribers.forEach(function (fn) {
          return fn();
        });
        state.current.subscribers.forEach(function (fn) {
          return fn(state.current);
        });
        if (renderFn) renderFn(state.current);else if (state.current.render && state.current.scene.children.length) {
          state.current.gl.render(state.current.scene, state.current.camera);
        }
      }
    }; // Start render-loop


    requestAnimationFrame(renderLoop); // Clean-up

    return function () {
      state.current.active = false;
      unmountComponentAtNode(state.current.scene);
    };
  }, []);
  React.useMemo(function () {
    if (ready) {
      state.current.gl.setSize(state.current.size.width, state.current.size.height, false);
      state.current.aspect = state.current.size.width / state.current.size.height;
      if (resize) resize(state.current);
      state.current.camera.aspect = state.current.aspect;
      state.current.camera.updateProjectionMatrix();
      state.current.camera.radius = (state.current.size.width + state.current.size.height) / 4;
    }
  }, [ready, state.current.size.width, state.current.size.height]);
  var intersect = React.useCallback(function (event, fn) {
    var x = event.clientX / state.current.size.width * 2 - 1;
    var y = -(event.clientY / state.current.size.height) * 2 + 1;
    mouse.set(x, y, 0.5);
    raycaster.setFromCamera(mouse, state.current.camera);
    var intersects = raycaster.intersectObjects(state.current.scene.children, true);

    for (var i = 0; i < intersects.length; i++) {
      var _intersect = intersects[i];
      if (!_intersect.object.__handlers) continue;
      fn(_intersect);
    }

    return intersects;
  }, []);
  React.useEffect(function () {
    var hovered = {};

    var handleMove = function handleMove(event) {
      var hover = false;
      var intersects = intersect(event, function (data) {
        var object = data.object;
        var handlers = object.__handlers;

        if (handlers.hover) {
          hover = true;

          if (!hovered[object.uuid]) {
            hovered[object.uuid] = object;
            handlers.hover(data);
          }
        }
      });
      if (hover) cursor !== 'pointer' && setCursor('pointer');else cursor !== 'default' && setCursor('default');
      Object.values(hovered).forEach(function (object) {
        if (!intersects.length || !intersects.find(function (i) {
          return i.object === object;
        })) {
          if (object.__handlers.unhover) {
            object.__handlers.unhover();
          }

          delete hovered[object.uuid];
        }
      });
    };

    var handleClick = function handleClick(event) {
      var clicked = {};
      intersect(event, function (data) {
        var object = data.object;
        var handlers = object.__handlers;

        if (handlers.click && !clicked[object.uuid]) {
          clicked[object.uuid] = object;
          handlers.click(data);
        }
      });
    };

    window.addEventListener('mousemove', handleMove, {
      passive: true
    });
    window.addEventListener('mouseup', handleClick, {
      passive: true
    });
    return function () {
      window.removeEventListener('mousemove', handleMove);
      window.removeEventListener('mouseup', handleClick);
    };
  });
  var IsReady = React.useCallback(function () {
    React.useEffect(function () {
      setReady(true);
      if (created) created(state.current);
    }, []);
    return null;
  }, []); // Render v-dom into scene

  React.useEffect(function () {
    if (state.current.size.width > 0) {
      render(React__default.createElement(context.Provider, {
        value: _extends({}, state.current)
      }, React__default.createElement(IsReady, null), typeof children === 'function' ? children(state.current) : children), state.current.scene);
    }
  }); // Render the canvas into the dom

  return React__default.createElement("div", _extends({}, bind, rest, {
    style: _extends({
      cursor: cursor,
      position: 'relative',
      width: '100%',
      height: '100%',
      overflow: 'hidden'
    }, style)
  }), React__default.createElement("canvas", {
    ref: canvas
  }));
});
function useRender(fn, main) {
  var _useContext = React.useContext(context),
      subscribe = _useContext.subscribe;

  React.useEffect(function () {
    return subscribe(fn, main);
  }, []);
}
function useThree(fn) {
  var _useContext2 = React.useContext(context),
      subscribe = _useContext2.subscribe,
      props = _objectWithoutPropertiesLoose(_useContext2, ["subscribe"]);

  return props;
}

exports.useFrameloop = useFrameloop;
exports.apply = apply;
exports.applyProps = applyProps;
exports.render = render;
exports.unmountComponentAtNode = unmountComponentAtNode;
exports.context = context;
exports.Canvas = Canvas;
exports.useRender = useRender;
exports.useThree = useThree;
