import _objectWithoutPropertiesLoose from '@babel/runtime/helpers/esm/objectWithoutPropertiesLoose';
import _extends from '@babel/runtime/helpers/esm/extends';
import * as THREE from 'three/src/Three';
import { Vector3, Math as Math$1, Raycaster, Vector2, WebGLRenderer, PerspectiveCamera, Scene } from 'three/src/Three';
import React, { useRef, useState, useEffect, useMemo, useCallback, useContext } from 'react';
import Reconciler from 'react-reconciler';
import omit from 'lodash-es/omit';
import upperFirst from 'lodash-es/upperFirst';
import ResizeObserver from 'resize-observer-polyfill';
import { unstable_now, unstable_scheduleCallback, unstable_cancelCallback } from 'scheduler';

const roots = new Map();
const emptyObject = {};
let globalSubscribersMap = new Set();
let globalSubscribers = [];
const useFrameloop = fn => {
  globalSubscribersMap.add(fn);
  globalSubscribers = Array.from(globalSubscribersMap.values());
};
let catalogue = {};
const apply = objects => catalogue = _extends({}, catalogue, objects);
function applyProps(instance, newProps, oldProps, interpolateArray) {
  if (oldProps === void 0) {
    oldProps = {};
  }

  if (interpolateArray === void 0) {
    interpolateArray = false;
  }

  if (instance.obj) instance = instance.obj; // Filter equals, events and reserved props

  const sameProps = Object.keys(newProps).filter(key => newProps[key] === oldProps[key]);
  const handlers = Object.keys(newProps).filter(key => typeof newProps[key] === 'function' && key.startsWith('on'));
  const filteredProps = omit(newProps, [...sameProps, 'children', 'key', 'ref']);

  if (Object.keys(filteredProps).length > 0) {
    Object.entries(filteredProps).forEach((_ref) => {
      let key = _ref[0],
          value = _ref[1];

      if (!handlers.includes(key)) {
        let root = instance;
        let target = root[key];

        if (key.includes('-')) {
          const entries = key.split('-');
          target = entries.reduce((acc, key) => acc[key], instance);

          if (target && !target.set) {
            // The target is atomic, this forces us to switch the root
            const _entries$reverse = entries.reverse(),
                  name = _entries$reverse[0],
                  reverseEntries = _entries$reverse.slice(1);

            root = reverseEntries.reverse().reduce((acc, key) => acc[key], instance);
            key = name;
          }
        }

        if (target && target.set) {
          if (target.constructor.name === value.constructor.name) {
            target.copy(value);
          } else if (Array.isArray(value)) {
            target.set(...value);
          } else {
            target.set(value);
          }
        } else {
          if (interpolateArray && Array.isArray(root[key])) root[key].push(value);else root[key] = value;
        }
      }
    }); // Prep interaction handlers

    if (handlers.length) {
      instance.__handlers = handlers.reduce((acc, key) => {
        const name = key.charAt(2).toLowerCase() + key.substr(3);
        return _extends({}, acc, {
          [name]: newProps[key]
        });
      }, {});
    } // Call the update lifecycle, if present


    if (instance.__handlers && instance.__handlers.update) instance.__handlers.update(instance);
  }
}

function createInstance(type, _ref2) {
  let _ref2$args = _ref2.args,
      args = _ref2$args === void 0 ? [] : _ref2$args,
      props = _objectWithoutPropertiesLoose(_ref2, ["args"]);

  let name = upperFirst(type);
  let instance;
  if (type === 'primitive') instance = props.object;else {
    const target = catalogue[name] || THREE[name];
    instance = Array.isArray(args) ? new target(...args) : new target(args);
  }
  if (!instance.isObject3D) instance = {
    obj: instance,
    parent: undefined
  };
  applyProps(instance, props, {});
  return instance;
}

function appendChild(parentInstance, child) {
  if (child) {
    if (child.isObject3D) parentInstance.add(child);else {
      child.parent = parentInstance; // The name attribute implies that the object attaches itself on the parent

      if (child.obj.name) applyProps(parentInstance, {
        [child.obj.name]: child.obj
      }, {}, true);
    }
  }
}

function removeChild(parentInstance, child) {
  if (child) {
    if (child.isObject3D) {
      parentInstance.remove(child);
      if (child.dispose) child.dispose();
    } else {
      child.parent = undefined;

      if (child.obj.name) {
        // Remove attachment
        const target = parentInstance[child.obj.name];
        if (Array.isArray(target)) parentInstance[child.obj.name] = target.filter(x => x !== child.obj);else parentInstance[child.obj.name] = undefined;
      }

      if (child.obj.dispose) child.obj.dispose();
    }
  }
}

const Renderer = Reconciler({
  now: unstable_now,
  createInstance,
  removeChild,
  appendChild,
  supportsMutation: true,
  isPrimaryRenderer: false,
  schedulePassiveEffects: unstable_scheduleCallback,
  cancelPassiveEffects: unstable_cancelCallback,
  appendInitialChild: appendChild,
  appendChildToContainer: appendChild,
  removeChildFromContainer: removeChild,

  insertBefore(parentInstance, child, beforeChild) {
    if (child) {
      if (child.isObject3D) {
        child.parent = parentInstance;
        child.dispatchEvent({
          type: 'added'
        }); // TODO: the order is out of whack if data objects are present, has to be recalculated

        const index = parentInstance.children.indexOf(beforeChild);
        parentInstance.children = [...parentInstance.children.slice(0, index), child, ...parentInstance.children.slice(index)];
      } else child.parent = parentInstance;
    }
  },

  commitUpdate(instance, updatePayload, type, oldProps, newProps, fiber) {
    if (instance.isObject3D) {
      applyProps(instance, newProps, oldProps);
    } else {
      // This is a data object, let's extract critical information about it
      const parent = instance.parent;

      const _newProps$args = newProps.args,
            argsNew = _newProps$args === void 0 ? [] : _newProps$args,
            restNew = _objectWithoutPropertiesLoose(newProps, ["args"]);

      const _oldProps$args = oldProps.args,
            argsOld = _oldProps$args === void 0 ? [] : _oldProps$args,
            restOld = _objectWithoutPropertiesLoose(oldProps, ["args"]); // If it has new props or arguments, then it needs to be re-instanciated


      if (argsNew.some((value, index) => value !== argsOld[index])) {
        // Next we create a new instance and append it again
        const newInstance = createInstance(instance.obj.type, newProps);
        removeChild(parent, instance);
        appendChild(parent, newInstance); // Switch instance

        instance.obj = newInstance.obj;
        instance.parent = newInstance.parent;
      } else {
        // Otherwise just overwrite props
        applyProps(instance, restNew, restOld);
      }
    }
  },

  getPublicInstance(instance) {
    return instance;
  },

  getRootHostContext(rootContainerInstance) {
    return emptyObject;
  },

  getChildHostContext(parentHostContext, type) {
    return emptyObject;
  },

  createTextInstance() {},

  finalizeInitialChildren(instance, type, props, rootContainerInstance) {
    return false;
  },

  prepareUpdate(instance, type, oldProps, newProps, rootContainerInstance, hostContext) {
    return emptyObject;
  },

  shouldDeprioritizeSubtree(type, props) {
    return false;
  },

  prepareForCommit() {},

  resetAfterCommit() {},

  shouldSetTextContent(props) {
    return false;
  }

});
function render(element, container) {
  let root = roots.get(container);

  if (!root) {
    root = Renderer.createContainer(container);
    roots.set(container, root);
  }

  Renderer.updateContainer(element, root, null, undefined);
  return Renderer.getPublicRootInstance(root);
}
function unmountComponentAtNode(container) {
  const root = roots.get(container);
  if (root) Renderer.updateContainer(null, root, null, () => roots.delete(container));
}

function useMeasure() {
  const ref = useRef();

  const _useState = useState({
    left: 0,
    top: 0,
    width: 0,
    height: 0
  }),
        bounds = _useState[0],
        set = _useState[1];

  const _useState2 = useState(() => new ResizeObserver((_ref3) => {
    let entry = _ref3[0];
    return set(entry.contentRect);
  })),
        ro = _useState2[0];

  useEffect(() => {
    if (ref.current) ro.observe(ref.current);
    return () => ro.disconnect();
  }, []);
  return [{
    ref
  }, bounds];
}

const context = React.createContext();
const Canvas = React.memo((_ref4) => {
  let children = _ref4.children,
      props = _ref4.props,
      style = _ref4.style,
      camera = _ref4.camera,
      renderFn = _ref4.render,
      resize = _ref4.resize,
      created = _ref4.created,
      rest = _objectWithoutPropertiesLoose(_ref4, ["children", "props", "style", "camera", "render", "resize", "created"]);

  const canvas = useRef();
  const state = useRef({
    subscribers: [],
    render: true,
    active: true,
    canvas: undefined,
    gl: undefined,
    camera: undefined,
    scene: undefined,
    size: undefined,
    viewport: function viewport(target) {
      if (target === void 0) {
        target = new Vector3(0, 0, 0);
      }

      const distance = state.current.camera.position.distanceTo(target);
      const fov = Math$1.degToRad(state.current.camera.fov); // convert vertical fov to radians

      const height = 2 * Math.tan(fov / 2) * distance; // visible height

      const width = height * state.current.camera.aspect;
      return {
        width,
        height
      };
    },
    subscribe: (fn, main) => {
      if (main !== void 0) state.current.render = main ? false : true;
      state.current.subscribers.push(fn);
      return () => state.current.subscribers = state.current.subscribers.filter(s => s === fn);
    }
  });

  const _useMeasure = useMeasure(),
        bind = _useMeasure[0],
        size = _useMeasure[1];

  state.current.size = size;

  const _useState3 = useState(false),
        ready = _useState3[0],
        setReady = _useState3[1];

  const readyRef = useRef(false);
  useEffect(() => void (readyRef.current = ready), [ready]);

  const _useState4 = useState(() => new Raycaster()),
        raycaster = _useState4[0];

  const _useState5 = useState(() => new Vector2()),
        mouse = _useState5[0];

  const _useState6 = useState('default'),
        cursor = _useState6[0],
        setCursor = _useState6[1];

  useEffect(() => {
    state.current.gl = new WebGLRenderer(_extends({
      canvas: canvas.current,
      antialias: true,
      alpha: true
    }, props));
    state.current.gl.setClearAlpha(0);
    state.current.gl.setSize(state.current.size.width, state.current.size.height, false); //state.current.gl.setPixelRatio(window.devicePixelRatio)
    //state.current.gl.setSize(window.innerWidth, window.innerHeight)

    state.current.camera = camera && camera.current || new PerspectiveCamera(75, 0, 0.1, 1000);
    state.current.camera.position.z = 5;
    state.current.canvas = canvas.current;
    state.current.scene = window.scene = new Scene();

    const renderLoop = function renderLoop() {
      if (!state.current.active) return;
      requestAnimationFrame(renderLoop);

      if (readyRef.current) {
        globalSubscribers.forEach(fn => fn());
        state.current.subscribers.forEach(fn => fn(state.current));
        if (renderFn) renderFn(state.current);else if (state.current.render && state.current.scene.children.length) {
          state.current.gl.render(state.current.scene, state.current.camera);
        }
      }
    }; // Start render-loop


    requestAnimationFrame(renderLoop); // Clean-up

    return () => {
      state.current.active = false;
      unmountComponentAtNode(state.current.scene);
    };
  }, []);
  useMemo(() => {
    if (ready) {
      state.current.gl.setSize(state.current.size.width, state.current.size.height, false);
      state.current.aspect = state.current.size.width / state.current.size.height;
      if (resize) resize(state.current);
      state.current.camera.aspect = state.current.aspect;
      state.current.camera.updateProjectionMatrix();
      state.current.camera.radius = (state.current.size.width + state.current.size.height) / 4;
    }
  }, [ready, state.current.size.width, state.current.size.height]);
  const intersect = useCallback((event, fn) => {
    const x = event.clientX / state.current.size.width * 2 - 1;
    const y = -(event.clientY / state.current.size.height) * 2 + 1;
    mouse.set(x, y, 0.5);
    raycaster.setFromCamera(mouse, state.current.camera);
    const intersects = raycaster.intersectObjects(state.current.scene.children, true);

    for (var i = 0; i < intersects.length; i++) {
      const intersect = intersects[i];
      if (!intersect.object.__handlers) continue;
      fn(intersect);
    }

    return intersects;
  }, []);
  useEffect(() => {
    const hovered = {};

    const handleMove = event => {
      let hover = false;
      let intersects = intersect(event, data => {
        const object = data.object;
        const handlers = object.__handlers;

        if (handlers.hover) {
          hover = true;

          if (!hovered[object.uuid]) {
            hovered[object.uuid] = object;
            handlers.hover(data);
          }
        }
      });
      if (hover) cursor !== 'pointer' && setCursor('pointer');else cursor !== 'default' && setCursor('default');
      Object.values(hovered).forEach(object => {
        if (!intersects.length || !intersects.find(i => i.object === object)) {
          if (object.__handlers.unhover) {
            object.__handlers.unhover();
          }

          delete hovered[object.uuid];
        }
      });
    };

    const handleClick = event => {
      const clicked = {};
      intersect(event, data => {
        const object = data.object;
        const handlers = object.__handlers;

        if (handlers.click && !clicked[object.uuid]) {
          clicked[object.uuid] = object;
          handlers.click(data);
        }
      });
    };

    window.addEventListener('mousemove', handleMove, {
      passive: true
    });
    window.addEventListener('mouseup', handleClick, {
      passive: true
    });
    return () => {
      window.removeEventListener('mousemove', handleMove);
      window.removeEventListener('mouseup', handleClick);
    };
  });
  const IsReady = useCallback(() => {
    useEffect(() => {
      setReady(true);
      if (created) created(state.current);
    }, []);
    return null;
  }, []); // Render v-dom into scene

  useEffect(() => {
    if (state.current.size.width > 0) {
      render(React.createElement(context.Provider, {
        value: _extends({}, state.current)
      }, React.createElement(IsReady, null), typeof children === 'function' ? children(state.current) : children), state.current.scene);
    }
  }); // Render the canvas into the dom

  return React.createElement("div", _extends({}, bind, rest, {
    style: _extends({
      cursor,
      position: 'relative',
      width: '100%',
      height: '100%',
      overflow: 'hidden'
    }, style)
  }), React.createElement("canvas", {
    ref: canvas
  }));
});
function useRender(fn, main) {
  const _useContext = useContext(context),
        subscribe = _useContext.subscribe;

  useEffect(() => subscribe(fn, main), []);
}
function useThree(fn) {
  const _useContext2 = useContext(context),
        subscribe = _useContext2.subscribe,
        props = _objectWithoutPropertiesLoose(_useContext2, ["subscribe"]);

  return props;
}

export { useFrameloop, apply, applyProps, render, unmountComponentAtNode, context, Canvas, useRender, useThree };
